@@ -1,4 +1,4 @@
-hurdle <- function(formula, data, subset, na.action,
+hurdle <- function(formula, data, subset, na.action, weights, offset,
                    dist = c("poisson", "negbin", "geometric"),
                    zero.dist = c("binomial", "poisson", "negbin", "geometric"),
                    link = c("logit", "probit", "cloglog", "cauchit", "log"),
@@ -8,59 +8,124 @@ hurdle <- function(formula, data, subset, na.action,
   ## set up likelihood components
   zeroPoisson <- function(parms) {
     ## mean
-    lambda <- as.vector(exp(Z %*% parms))
+    mu <- as.vector(exp(Z %*% parms))
     ## log-likelihood
-    loglik0 <- -lambda ## = dpois(0, lambda, log = TRUE)
+    loglik0 <- -mu ## = dpois(0, lambda = mu, log = TRUE)
     ## collect and return
-    loglik <- sum(loglik0[Y0]) + sum(log(1 - exp(loglik0[Y1])))
+    loglik <- sum(weights[Y0] * loglik0[Y0]) + sum(weights[Y1] * log(1 - exp(loglik0[Y1])))
     loglik
   }
 
   countPoisson <- function(parms) {
     ## mean
-    lambda <- as.vector(exp(X %*% parms))[Y1]
+    mu <- as.vector(exp(X %*% parms + offset))[Y1]
     ## log-likelihood
-    loglik0 <- -lambda ## = dpois(0, lambda, log = TRUE)
-    loglik1 <- dpois(Y[Y1], lambda, log = TRUE)
+    loglik0 <- -mu ## = dpois(0, lambda = mu, log = TRUE)
+    loglik1 <- dpois(Y[Y1], lambda = mu, log = TRUE)
     ## collect and return
-    loglik <- sum(loglik1) - sum(log(1 - exp(loglik0)))
+    loglik <- sum(weights[Y1] * loglik1) - sum(weights[Y1] * log(1 - exp(loglik0)))
     loglik
   }
 
   zeroNegBin <- function(parms) {
     ## parameters
-    lambda <- as.vector(exp(Z %*% parms[1:kz]))
+    mu <- as.vector(exp(Z %*% parms[1:kz]))
     theta <- exp(parms[kz+1])
     ## log-likelihood
-    loglik0 <- suppressWarnings(dnbinom(0, size = theta, mu = lambda, log = TRUE))
+    loglik0 <- suppressWarnings(dnbinom(0, size = theta, mu = mu, log = TRUE))
     ## collect and return
-    loglik <- sum(loglik0[Y0]) + sum(log(1 - exp(loglik0[Y1])))
+    loglik <- sum(weights[Y0] * loglik0[Y0]) + sum(weights[Y1] * log(1 - exp(loglik0[Y1])))
     loglik
   }
 
   countNegBin <- function(parms) {
     ## parameters
-    lambda <- as.vector(exp(X %*% parms[1:kx]))[Y1]
+    mu <- as.vector(exp(X %*% parms[1:kx] + offset))[Y1]
     theta <- exp(parms[kx+1])
     ## log-likelihood
-    loglik0 <- suppressWarnings(dnbinom(0, size = theta, mu = lambda, log = TRUE))
-    loglik1 <- suppressWarnings(dnbinom(Y[Y1], size = theta, mu = lambda, log = TRUE))
+    loglik0 <- suppressWarnings(dnbinom(0, size = theta, mu = mu, log = TRUE))
+    loglik1 <- suppressWarnings(dnbinom(Y[Y1], size = theta, mu = mu, log = TRUE))
     ## collect and return
-    loglik <- sum(loglik1) - sum(log(1 - exp(loglik0)))
+    loglik <- sum(weights[Y1] * loglik1) - sum(weights[Y1] * log(1 - exp(loglik0)))
     loglik
   }
 
   zeroGeom <- function(parms) zeroNegBin(c(parms, 0))
+  
   countGeom <- function(parms) countNegBin(c(parms, 0))
 
   zeroBinom <- function(parms) {
     ## mean
-    lambda <- as.vector(linkinv(Z %*% parms))
+    mu <- as.vector(linkinv(Z %*% parms))
     ## log-likelihood
-    loglik <- sum(log(1 - lambda[Y0])) + sum(log(lambda[Y1]))
+    loglik <- sum(weights[Y0] * log(1 - mu[Y0])) + sum(weights[Y1] * log(mu[Y1]))
     loglik  
   }
 
+  countGradPoisson <- function(parms) {
+    eta <- as.vector(X %*% parms + offset)[Y1]
+    mu <- exp(eta)
+    colSums(((Y[Y1] - mu) - exp(ppois(0, lambda = mu, log.p = TRUE) -
+      ppois(0, lambda = mu, lower.tail = FALSE, log.p = TRUE) + eta)) * weights[Y1] * X[Y1, , drop = FALSE])
+  }
+  
+  countGradGeom <- function(parms) {
+    eta <- as.vector(X %*% parms + offset)[Y1]
+    mu <- exp(eta)      
+    colSums(((Y[Y1] - mu * (Y[Y1] + 1)/(mu + 1)) -
+      exp(pnbinom(0, mu = mu, size = 1, log.p = TRUE) -
+        pnbinom(0, mu = mu, size = 1, lower.tail = FALSE, log.p = TRUE) -
+	log(mu + 1) + eta)) * weights[Y1] * X[Y1, , drop = FALSE])
+  }
+
+  countGradNegBin <- function(parms) {
+    eta <- as.vector(X %*% parms[1:kx] + offset)[Y1]
+    mu <- exp(eta)      
+    theta <- exp(parms[kx+1])
+    logratio <- pnbinom(0, mu = mu, size = theta, log.p = TRUE) -
+        pnbinom(0, mu = mu, size = theta, lower.tail = FALSE, log.p = TRUE)
+    rval <- colSums(((Y[Y1] - mu * (Y[Y1] + theta)/(mu + theta)) -
+      exp(logratio + log(theta) - log(mu + theta) + eta)) * weights[Y1] * X[Y1, , drop = FALSE])
+    rval2 <- sum((digamma(Y[Y1] + theta) - digamma(theta) +    
+      log(theta) - log(mu + theta) + 1 - (Y[Y1] + theta)/(mu + theta) +
+      exp(logratio) * (log(theta) - log(mu + theta) + 1 - theta/(mu + theta))) * weights[Y1]) * theta
+    c(rval, rval2)
+  }  
+  
+  zeroGradPoisson <- function(parms) {
+    eta <- as.vector(Z %*% parms)
+    mu <- exp(eta)
+    colSums(ifelse(Y0, -mu, exp(ppois(0, lambda = mu, log.p = TRUE) -
+      ppois(0, lambda = mu, lower.tail = FALSE, log.p = TRUE) + eta)) * weights * Z)
+  }
+
+  zeroGradGeom <- function(parms) {
+    eta <- as.vector(Z %*% parms)
+    mu <- exp(eta)
+    colSums(ifelse(Y0, -mu/(mu + 1), exp(pnbinom(0, mu = mu, size = 1, log.p = TRUE) -
+      pnbinom(0, mu = mu, size = 1, lower.tail = FALSE, log.p = TRUE) - log(mu + 1) + eta)) * weights * Z)
+  }
+
+  zeroGradNegBin <- function(parms) {
+    eta <- as.vector(Z %*% parms[1:kz])
+    mu <- exp(eta)
+    theta <- exp(parms[kz+1])
+    logratio <- pnbinom(0, mu = mu, size = theta, log.p = TRUE) -
+      pnbinom(0, mu = mu, size = theta, lower.tail = FALSE, log.p = TRUE)
+    rval <- colSums(ifelse(Y0, -mu * theta/(mu + theta),
+      exp(logratio + log(theta) - log(mu + theta) + eta)) * weights * Z)
+    rval2 <- sum(ifelse(Y0, log(theta) - log(mu + theta) + 1 - theta/(mu + theta),
+      -exp(logratio) * (log(theta) - log(mu + theta) + 1 - theta/(mu + theta))) * weights * theta)
+    c(rval, rval2)
+  }
+
+  zeroGradBinom <- function(parms) {
+    eta <- as.vector(Z %*% parms)
+    mu <- linkinv(eta)
+    colSums(ifelse(Y0, -1/(1-mu), 1/mu)  * linkobj$mu.eta(eta) * weights * Z)  
+  }
+
+  ## collect likelihood components
   dist <- match.arg(dist)
   zero.dist <- match.arg(zero.dist)
   countDist <- switch(dist,
@@ -72,16 +137,29 @@ hurdle <- function(formula, data, subset, na.action,
 		     "geometric" = zeroGeom,
 		     "negbin" = zeroNegBin,
 		     "binomial" = zeroBinom)
+  countGrad <- switch(dist,
+                     "poisson" = countGradPoisson,
+		     "geometric" = countGradGeom,
+		     "negbin" = countGradNegBin)
+  zeroGrad <- switch(zero.dist,
+                     "poisson" = zeroGradPoisson,
+		     "geometric" = zeroGradGeom,
+		     "negbin" = zeroGradNegBin,
+		     "binomial" = zeroGradBinom)
   loglikfun <- function(parms) countDist(parms[1:(kx + (dist == "negbin"))]) +
     zeroDist(parms[(kx + (dist == "negbin") + 1):(kx + kz + (dist == "negbin") + (zero.dist == "negbin"))])
+  gradfun <- function(parms) c(countGrad(parms[1:(kx + (dist == "negbin"))]),
+    zeroGrad(parms[(kx + (dist == "negbin") + 1):(kx + kz + (dist == "negbin") + (zero.dist == "negbin"))]))
+
 
   ## binary link processing
-  link <- match.arg(link)
-  linkinv <- make.link(link)$linkinv
+  linkstr <- match.arg(link)
+  linkobj <- make.link(linkstr)
+  linkinv <- linkobj$linkinv
 
   if(control$trace) cat("Hurdle Count Model\n",
     paste("count model:", dist, "with log link\n"),
-    paste("zero hurdle model:", zero.dist, "with", ifelse(zero.dist == "binomial", link, "log"), "link\n"),
+    paste("zero hurdle model:", zero.dist, "with", ifelse(zero.dist == "binomial", linkstr, "log"), "link\n"),
     sep = "")
 
 
@@ -89,7 +167,7 @@ hurdle <- function(formula, data, subset, na.action,
   cl <- match.call()
   if(missing(data)) data <- environment(formula)
   mf <- match.call(expand.dots = FALSE)
-  m <- match(c("formula", "data", "subset", "na.action"), names(mf), 0)
+  m <- match(c("formula", "data", "subset", "na.action", "weights", "offset"), names(mf), 0)
   mf <- mf[c(1, m)]
   mf$drop.unused.levels <- TRUE
   
@@ -146,9 +224,21 @@ hurdle <- function(formula, data, subset, na.action,
   n <- length(Y)
   kx <- NCOL(X)
   kz <- NCOL(Z)
-  Y0 <- Y < 1
+  Y0 <- Y <= 0
   Y1 <- Y > 0
-    
+
+  ## weights and offset
+  weights <- model.weights(mf)
+  if(is.null(weights)) weights <- 1
+  if(length(weights) == 1) weights <- rep(weights, n)
+  weights <- as.vector(weights)
+  names(weights) <- rownames(mf)
+  
+  offset <- model.offset(mf)
+  if(is.null(offset)) offset <- 0
+  if(length(offset) == 1) offset <- rep(offset, n)
+  offset <- as.vector(offset)
+
   ## starting values
   start <- control$start
   if(!is.null(start)) {
@@ -185,12 +275,12 @@ hurdle <- function(formula, data, subset, na.action,
   
   if(is.null(start)) {
     if(control$trace) cat("generating starting values...")
-    model_count <- glm.fit(X, Y, family = poisson())
+    model_count <- glm.fit(X, Y, family = poisson(), weights = weights, offset = offset)
     model_zero <- switch(zero.dist,
-      "poisson" = glm.fit(Z, Y, family = poisson()),
-      "negbin" = glm.fit(Z, Y, family = poisson()),
-      "geometric" = suppressWarnings(glm.fit(Z, factor(Y > 0), family = binomial())),
-      "binomial" = suppressWarnings(glm.fit(Z, factor(Y > 0), family = binomial(link = link))))
+      "poisson" = glm.fit(Z, Y, family = poisson(), weights = weights),
+      "negbin" = glm.fit(Z, Y, family = poisson(), weights = weights),
+      "geometric" = suppressWarnings(glm.fit(Z, factor(Y > 0), family = binomial(), weights = weights)),
+      "binomial" = suppressWarnings(glm.fit(Z, factor(Y > 0), family = binomial(link = linkstr), weights = weights)))
     start <- list(count = model_count$coefficients, zero = model_zero$coefficients)
     start$theta <- c(count = if(dist == "negbin") 1 else NULL,
                      zero = if(zero.dist == "negbin") 1 else NULL)
@@ -209,12 +299,12 @@ hurdle <- function(formula, data, subset, na.action,
   ## separate estimation of censored and truncated component...
   if(separate) {
     if(control$trace) cat("calling optim() for count component estimation:\n")
-    fit_count <- optim(fn = countDist,
+    fit_count <- optim(fn = countDist, gr = countGrad,
       par = c(start$count, if(dist == "negbin") log(start$theta["count"]) else NULL),
       method = method, hessian = hessian, control = control)
 
     if(control$trace) cat("calling optim() for zero hurdle component estimation:\n")
-    fit_zero <- optim(fn = zeroDist,
+    fit_zero <- optim(fn = zeroDist, gr = zeroGrad,
       par = c(start$zero,  if(zero.dist == "negbin") log(start$theta["zero"]) else NULL),
       method = method, hessian = hessian, control = control)
     if(control$trace) cat("done\n")
@@ -242,7 +332,7 @@ hurdle <- function(formula, data, subset, na.action,
   } else {
   ## ...or joint.
     if(control$trace) cat("calling optim() for joint count and zero hurlde estimation:\n")
-    fit <- optim(fn = loglikfun,
+    fit <- optim(fn = loglikfun, gr = gradfun,
       par = c(start$count, if(dist == "negbin") log(start$theta["count"]) else NULL,
               start$zero,  if(zero.dist == "negbin") log(start$theta["zero"]) else NULL),
       method = method, hessian = hessian, control = control)
@@ -276,18 +366,18 @@ hurdle <- function(formula, data, subset, na.action,
   ## fitted and residuals
   phi <- if(zero.dist == "binomial") linkinv(Z %*% coefz)[,1] else exp(Z %*% coefz)[,1]
   p0_zero <- switch(zero.dist,
-                    "binomial" = 1-phi,
-		    "poisson" = exp(-phi),
-		    "negbin" = dnbinom(0, size = theta["zero"], mu = phi),
-		    "geometric" = dnbinom(0, size = 1, mu = phi))
+  		    "binomial" = log(phi),
+        	    "poisson" = ppois(0, lambda = phi, lower.tail = FALSE, log.p = TRUE),
+        	    "negbin" = pnbinom(0, size = theta["zero"], mu = phi, lower.tail = FALSE, log.p = TRUE),
+        	    "geometric" = pnbinom(0, size = 1, mu = phi, lower.tail = FALSE, log.p = TRUE))
 
-  mu <- exp(X %*% coefc)[,1]
+  mu <- exp(X %*% coefc + offset)[,1]
   p0_count <- switch(dist,
-		    "poisson" = exp(-mu),
-		    "negbin" = dnbinom(0, size = theta["count"], mu = mu),
-		    "geometric" = dnbinom(0, size = 1, mu = mu))
+        	    "poisson" = ppois(0, lambda = mu, lower.tail = FALSE, log.p = TRUE),
+        	    "negbin" = pnbinom(0, size = theta["count"], mu = mu, lower.tail = FALSE, log.p = TRUE),
+        	    "geometric" = pnbinom(0, size = 1, mu = mu, lower.tail = FALSE, log.p = TRUE))
+  Yhat <- exp((p0_zero - p0_count) + log(mu))
 
-  Yhat <- (1 - p0_zero) * mu / (1 - p0_count)
   res <- Y - Yhat
 
 
@@ -298,16 +388,18 @@ hurdle <- function(formula, data, subset, na.action,
     method = method,
     control = control,
     start = start,
+    weights = weights,
+    offset = if(identical(offset, rep(0, n))) NULL else offset,
     n = n,
     df.null = n - 2,
-    df.residual = n - (kx + kz + 2 * (dist == "negbin")),
+    df.residual = n - (kx + kz + (dist == "negbin") + (zero.dist == "negbin")),
     terms = list(count = mtX, zero = mtZ, full = mt),
     theta = theta,
     SE.logtheta = SE.logtheta,
     loglik = if(separate) fit_count$value + fit_zero$value else fit$value,
     vcov = vc,
     dist = list(count = dist, zero = zero.dist),
-    link = if(zero.dist == "binomial") link else NULL,
+    link = if(zero.dist == "binomial") linkstr else NULL,
     linkinv = if(zero.dist == "binomial") linkinv else NULL,
     separate = separate,
     converged = if(separate) fit_count$convergence < 1 & fit_zero$convergence < 1 else fit$convergence < 1,
@@ -461,13 +553,14 @@ model.matrix.hurdle <- function(object, model = c("count", "zero"), ...) {
   return(rval)
 }
 
-predict.hurdle <- function(object, newdata, type = c("response", "prob"), na.action = na.pass, ...)
+predict.hurdle <- function(object, newdata, type = c("response", "prob", "count", "zero"),
+  na.action = na.pass, ...)
 {
     type <- match.arg(type)
 
     ## if no new data supplied
     if(missing(newdata)) {
-      if(type == "prob") {
+      if(type != "response") {
         if(!is.null(object$x)) {
 	  X <- object$x$count
 	  Z <- object$x$zero
@@ -477,6 +570,7 @@ predict.hurdle <- function(object, newdata, type = c("response", "prob"), na.act
 	} else {
 	  stop("predicted probabilities cannot be computed with missing newdata")
 	}
+	offset <- if(is.null(object$offset)) rep(0, NROW(X)) else object$offset
       } else {
         return(object$fitted.values)
       }
@@ -484,23 +578,31 @@ predict.hurdle <- function(object, newdata, type = c("response", "prob"), na.act
       mf <- model.frame(delete.response(object$terms$full), newdata, na.action = na.action, xlev = object$levels)
       X <- model.matrix(delete.response(object$terms$count), mf, contrasts = object$contrasts$count)
       Z <- model.matrix(delete.response(object$terms$zero),  mf, contrasts = object$contrasts$zero)
+      offset <- if(!is.null(off.num <- attr(object$terms$full, "offset"))) 
+          eval(attr(object$terms$full, "variables")[[off.num + 1]], newdata)
+        else if(!is.null(object$offset)) eval(object$call$offset, newdata)
+      if(is.null(offset)) offset <- rep(0, NROW(X))
     }
 
     phi <- if(object$dist$zero == "binomial") object$linkinv(Z %*% object$coefficients$zero)[,1]
       else exp(Z %*% object$coefficients$zero)[,1]
     p0_zero <- switch(object$dist$zero,
-                      "binomial" = 1-phi,
-		      "poisson" = exp(-phi),
-		      "negbin" = dnbinom(0, size = object$theta["zero"], mu = phi),
-		      "geometric" = dnbinom(0, size = 1, mu = phi))
+                      "binomial" = log(phi),
+		      "poisson" = ppois(0, lambda = mu, lower.tail = FALSE, log.p = TRUE),
+		      "negbin" = pnbinom(0, size = object$theta["zero"], mu = phi, lower.tail = FALSE, log.p = TRUE),
+		      "geometric" = pnbinom(0, size = 1, mu = phi, lower.tail = FALSE, log.p = TRUE))
 
-    mu <- exp(X %*% object$coefficients$count)[,1]
+    mu <- exp(X %*% object$coefficients$count + offset)[,1]
     p0_count <- switch(object$dist$count,
-		      "poisson" = exp(-mu),
-		      "negbin" = dnbinom(0, size = object$theta["count"], mu = mu),
-		      "geometric" = dnbinom(0, size = 1, mu = mu))
-
+		      "poisson" = ppois(0, lambda = mu, lower.tail = FALSE, log.p = TRUE),
+		      "negbin" = pnbinom(0, size = object$theta["count"], mu = mu, lower.tail = FALSE, log.p = TRUE),
+		      "geometric" = pnbinom(0, size = 1, mu = mu, lower.tail = FALSE, log.p = TRUE))
+    logphi <- p0_zero - p0_count
     
+    if(type == "response") rval <- exp(logphi + log(mu))
+    if(type == "count") rval <- mu
+    if(type == "zero") rval <- exp(logphi)
+
     ## predicted probabilities
     if(type == "prob") {
       if(!is.null(object$y)) y <- object$y
@@ -512,19 +614,17 @@ predict.hurdle <- function(object, newdata, type = c("response", "prob"), na.act
       rval <- matrix(NA, nrow = length(mu), ncol = nUnique)
       dimnames(rval) <- list(rownames(X), yUnique)
       
-      rval[,1] <- p0_zero
+      rval[,1] <- 1 - exp(p0_zero)
       switch(object$dist$count,
              "poisson" = {
-               for(i in 2:nUnique) rval[,i] <- (1-p0_zero)/(1-p0_count) * dpois(yUnique[i], lambda = mu)
+               for(i in 2:nUnique) rval[,i] <- exp(logphi + dpois(yUnique[i], lambda = mu, log = TRUE))
 	     },
 	     "negbin" = {
-               for(i in 2:nUnique) rval[,i] <- (1-p0_zero)/(1-p0_count) * dnbinom(yUnique[i], mu = mu, size = object$theta["count"])
+               for(i in 2:nUnique) rval[,i] <- exp(logphi + dnbinom(yUnique[i], mu = mu, size = object$theta["count"], log = TRUE))
 	     },
 	     "geometric" = {
-               for(i in 2:nUnique) rval[,i] <- (1-p0_zero)/(1-p0_count) * dnbinom(yUnique[i], mu = mu, size = 1)
+               for(i in 2:nUnique) rval[,i] <- exp(logphi + dnbinom(yUnique[i], mu = mu, size = 1, log = TRUE))
 	     })
-    } else {
-      rval <- (1 - p0_zero) * mu / (1 - p0_count)
     }
    
     rval
@@ -536,12 +636,34 @@ fitted.hurdle <- function(object, ...) {
 
 residuals.hurdle <- function(object, type = c("pearson", "response"), ...) {
   type <- match.arg(type)
-  if(type == "pearson") return(object$residuals/sqrt(object$fitted.values))
-    else return(object$residuals)
+  if(type == "response") return(object$residuals) else {
+    mu <- predict(object, type = "count")
+    phi <- predict(object, type = "zero")
+    theta1 <- switch(object$dist$count,
+      "poisson" = 0,
+      "geometric" = 1,
+      "negbin" = 1/object$theta)
+    vv <- object$fitted.values * (1 + ((1-phi) + theta1) * mu)
+    return(object$residuals/sqrt(vv))  
+  }  
 }
 
 predprob.hurdle <- function(obj, ...){
     predict(obj, type = "prob", ...)
 }
 
-## estfun.hurdle?
+extractAIC.hurdle <- function(fit, scale = NULL, k = 2, ...) {
+  c(attr(logLik(fit), "df"), AIC(fit, k = k))
+}
+
+hurdletest <- function(object, ...) {
+  stopifnot(inherits(object, "hurdle"))
+  stopifnot(object$dist$count == object$dist$zero)
+  stopifnot(all(sort(names(object$coefficients$count)) == sort(names(object$coefficients$zero))))
+  stopifnot(require("car"))
+  nam <- names(object$coefficients$count)
+  lh <- paste("count_", nam, " = ", "zero_", nam, sep = "")
+  rval <- car::linear.hypothesis(object, lh, ...)
+  attr(rval, "heading")[1] <- "Wald test for hurdle models\n\nRestrictions:"
+  return(rval)
+}
